use crate::result::{Context, Result};

pub struct Aws;

#[derive(Debug)]
pub struct InvalidHeartbeatResponse(u8);

impl std::fmt::Display for InvalidHeartbeatResponse {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "Invalid heartbeat response: expected 0x87, got 0x{}",
            self.0
        )
    }
}

impl std::error::Error for InvalidHeartbeatResponse {}

impl Aws {
    fn init_heartbeat() -> Result<()> {
        use crate::system::syscall::{
            close, connect, read, sockaddr_vm, socket, write, SocketFamily, SocketType,
        };

        let mut sockaddr = sockaddr_vm {
            svm_family: SocketFamily::Vsock as u16,
            svm_reserved1: 0,
            svm_port: 9000,
            svm_cid: 3,
            svm_zero: [0; 4],
        };
        let mut buf = [0x87];

        let fd = socket(SocketFamily::Vsock, SocketType::Stream)?;
        unsafe {
            connect(
                fd,
                std::ptr::from_mut(&mut sockaddr).cast(),
                std::mem::size_of_val(&sockaddr),
            )?;
        };
        write(fd, &buf)?;
        read(fd, &mut buf)?;
        close(fd)?;

        if buf[0] != 0x87 {
            return Err(InvalidHeartbeatResponse(buf[0]))
                .context(format_args!("Bad value from heartbeat"));
        }

        Ok(())
    }
}

impl super::Platform for Aws {
    fn is(&self) -> Result<bool> {
        std::fs::exists("/dev/nsm").context(format_args!("could not check if /dev/nsm exists"))
    }

    fn get_modules(&self) -> Result<Vec<(String, String)>> {
        Ok(vec![("/nsm.ko".into(), String::new())])
    }

    fn init(&self) -> Result<()> {
        Self::init_heartbeat()?;
        Ok(())
    }
}
